/************************************************/
// Genetic Trail (Evolution Simulation)
// Author: Ilya Grigorik (2005)
// Site: http://igvita.com/blog/?page_id=68
/************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Diagnostics;

namespace GeneticTrail
{
    class Program
    {
        static int Main(string[] args)
        {

            if (args.Length < 5 || args.Length > 5) 
            {
                Console.WriteLine("Usage: GeneticTrail <Trail Number> <Population Size> <genotype Length> <mutation Rate> <moves>");
                return 1;
            }

            bool flag = false;

            try
            {
                // Got the arguments, conjure up Adam and Eve and all that other 'stuff'.
                // After they eat that damn apple just drop them them on earth, press the evolution button and forget about them forever.
                // P.S. If things get real messed up, send in jebus at some point.
                Population p = new Population(int.Parse(args[1]), int.Parse(args[2]), int.Parse(args[3]), int.Parse(args[0]));
                for (int idx = 0; !flag ; idx++)
                {
                    flag = p.runGeneration(int.Parse(args[4]));
                    if (!flag) p.advanceGeneration();
                }
            }
            catch (System.FormatException)
            {
                Console.WriteLine("Please enter numeric arguments... What kind of God are you? Sheesh.");
                Console.WriteLine("Usage: GeneticTrail <Trail Number> <Population Size> <genotype Length> <mutation Rate> <moves>");
                return 1;
            }

            return 0;
        }
    }

    class Trail
    {
        public ArrayList path;

        public Trail(int num)
        {
            path = new ArrayList();
            int[] steps;
            
            // Define path through the 10x10 matrix 
            //  -- Only two worlds for now.
            //    #1 - 35 steps
            //    #2 - 19 steps

            if (num == 1)
            {
                steps = new int[]  { 91, 81, 82, 72, 73, 83, 84, 85, 75, 65, 55, 45,
                                     44, 43, 33, 32, 31, 21, 11, 12, 13, 14, 15, 16, 
                                     26, 27, 37, 47, 48, 49, 39, 29, 30, 20, 10 };
            }
            else
            {
                steps = new int[] {91, 81, 82, 72, 73, 74, 64, 54, 44, 45, 46,
                                   36, 37, 38, 28, 18, 8, 9, 10 };
            }

            path.AddRange(steps);
        }

        public bool checkTrail(int loc)
        {
            return path.Contains(loc);
        }

        public int Size()
        {
            return path.Count;
        }
   }

    class Population
    {
        ArrayList population;
        int populationSize;
        int genotypeLength;
        int mutateRate;
        int genNumber = 1;
        Random r = new Random();
        Trail p;

        public Population(int pSize, int gLength, int mRate, int tNum)
        {
            // Do all the stuff that took us 7 days last time in here... 
            populationSize = pSize;
            genotypeLength = gLength;
            mutateRate = mRate;
            population = new ArrayList();
            p = new Trail(tNum);
            
            // Generate the initial population with random genotype for each 'subject'.
            for (int idx = 0; idx < populationSize; idx++)
                population.Add(new Genotype(genotypeLength, r, p));

            Console.WriteLine();
            Console.WriteLine("Population statistics:");
            Console.WriteLine(("").PadRight(30, '-'));
            Console.WriteLine("Population Size:\t{0}", populationSize);
            Console.WriteLine("Phenotype Length:\t{0}", genotypeLength);
            Console.WriteLine("Mutation Rate:\t\t{0}", mRate);  
        }

        public bool runGeneration(int moves) {
            GenerationStats currGen = new GenerationStats();

            // Let'em sinners go at it here.. Maybe one of them will do something good for once. (aka run the simulation)
            foreach (Genotype gen in population)
            {     
                int fit = gen.Explore(moves, r);

                if (fit == p.Size()) { // Well what do you know. Someone tortured himself enough to get admitted back to heaven. 
                                       // -- We found a solution, exit happily. (Run apocalypse() before finishing.) 
                    Console.Beep();
                    Console.WriteLine(("").PadRight(40, '-')); 
                    Console.WriteLine("Solution Found");
                    Console.WriteLine(("").PadRight(40, '-'));
                    gen.printGene();
                    return true; 
                }

                else currGen.add(fit);
            }

            int avgFit = currGen.calculateStats();   // Calculate stats for this generation (Average & Std. Dev) 

            foreach (Genotype gen in population)     // Update likelihoods for cross-breeding.
                gen.calcLikelihood(populationSize, avgFit);

            // Sort the population by fitness, let the fit males go at it first ;)
            population.Sort();

            //foreach (Genotype gen in population)   // Omnipotent and all.. Debugging helps sometimes too.
            //   gen.printGene();

            // Print generation stats
            currGen.printStats(genNumber);
            return false;
        }

        public void advanceGeneration()
        {
            ArrayList newPopulation = new ArrayList();
            
            // Roulette Wheel Selection: Choose father/mother and let them indulge in earthly pleasures.
            ArrayList decisionVec = new ArrayList();
            int currGene = 0;

            //  1. Create a decision vector based on likelihoods
            foreach (Genotype gen in population)
            {
                for (double idx = 0; (idx <= Math.Round(gen.likelihood,3) ) && (decisionVec.Count < populationSize/2) ; idx+=0.005)
                       decisionVec.Add(currGene);

                if (decisionVec.Count >= populationSize/2) break;
                currGene++;
            }

            // Print decision vector
            //System.Collections.IEnumerator myEnumerator = decisionVec.GetEnumerator();
            //while (myEnumerator.MoveNext())
            //    Console.Write("{0} ", myEnumerator.Current);
            //Console.WriteLine();

            //  2. Pick two random parents and do a crossover + mutate
            //     -- To avoid premature convergence, mutate the genes to add some variation to this whole hooplah.
            for (int idx = 0; idx < populationSize; idx++)
            {
                int father = 0, mother = 0;
                while (father == mother) // Don't breed with yourself.. That's just wrong. 
                {
                    father = r.Next(0, decisionVec.Count-1);
                    mother = r.Next(0, decisionVec.Count-1);
                }

                // Crossover the genotypes. *Soft music in the background*
                Genotype newGene = crossOver((int)decisionVec[father], (int)decisionVec[mother]); 
                
                // Mutate the gene (Using same decision vector & mutationRate as the cutoff)
                if (r.Next(0, 100) <= mutateRate) newGene.mutate(r);

                // Add to the new population
                newPopulation.Add(newGene);
            }

            population = newPopulation;
            genNumber++;
        }

        private Genotype crossOver(int fatherId, int motherId)
        {
            //Console.WriteLine("Crossing {0} with {1}", fatherId, motherId);
            
            // Breed a new specie here!
            // Simple One-Point Crossover
            ArrayList gene1, gene2;
            
            int splitPoint = r.Next(0, genotypeLength - 1);
            int order = r.Next(1, 2);

            Genotype father = (Genotype)population[fatherId];
            Genotype mother = (Genotype)population[motherId];

            if (order == 1)
            {
                gene1 = father.retrievePhenotype(splitPoint, true);
                gene2 = mother.retrievePhenotype(splitPoint, false);
            }
            else
            {
                gene1 = mother.retrievePhenotype(splitPoint, true);
                gene2 = father.retrievePhenotype(splitPoint, false);
            }

            return new Genotype(gene1, gene2, p);
        }
    }

    class GenerationStats
    {
        // Keep some stats about all the sinners in here. 
        //  -- Even a god needs a database class to keep track of good and bad deeds...

        ArrayList generation = new ArrayList();
        int maxFit = 0;
        int lowFit = 100;
        int total = 0;
        int avgFitness = 0;
        double stdDev;
        double doubleAvg;

        public void add(int fitness)
        {
            generation.Add(fitness);

            if (fitness > maxFit) maxFit = fitness;
            if (fitness < lowFit) lowFit = fitness;

            total += fitness;
        }

        public double GetAvg(ArrayList num)
        {
            // Calculate the average sinner's IQ in here. For book keeping purposes 
            // and to help decide when things get out of hand and Jebus needs to be sent.

            double sum = 0.0;
            for (int i = 0; i < num.Count; i++)
            {
                sum += System.Convert.ToDouble(num[i]);
            }
            double avg = sum / System.Convert.ToDouble(num.Count);

            return avg;
        }

        public double GetStandardDeviation(ArrayList num)
        {
            // God is a math geek, didn't you know? Calculate the 
            // standard deviation of the distribution within the population.

            double SumOfSqrs = 0;
            doubleAvg = GetAvg(num);
            for (int i = 0; i < num.Count; i++)
            {
                SumOfSqrs += Math.Pow((Convert.ToDouble(num[i]) - doubleAvg), 2);
            }
            double n = (double)num.Count;
            return Math.Sqrt(SumOfSqrs / (n - 1));
        }

        public int calculateStats()
        {
            avgFitness = total / generation.Count;
            stdDev = GetStandardDeviation(generation);
            return avgFitness;
        }

        public void printStats(int gNum)
        {
            Console.WriteLine();
            Console.WriteLine("Generation #{0}", gNum);
            Console.WriteLine(("").PadRight(60, '-'));
            Console.WriteLine("Max Fitness:\t{0}\tMin Fitness:\t{1}", maxFit, lowFit);
            Console.WriteLine("Avg. Fitness:\t{0}\tStd.Deviation:\t{1}", doubleAvg, stdDev.ToString("N10")); 

         }

    }


    class Genotype : System.IComparable
    {
        // Recipe: 
        //  1) Anesthesize yourself, or just get really drunk. 
        //  2) Take own rib and mold something out of it. Call it 'something'.
        //  3) Stick a genotype into it and a bunch of other useless things. 

        Trail trailPath;
        ArrayList phenotype;
        int fitness;   
        int location; 
        int lastLocation;
        public double likelihood;

        public Genotype(int length, Random r, Trail p)
        {
            // Define a random genotype - Include 91 as start point. 
            phenotype = new ArrayList();
            phenotype.Add(91); // startpoint
            trailPath = p;

            for (int i = 1; i <= length-1; i++)
                phenotype.Add(r.Next(1, 100));

        }

        public Genotype(ArrayList gene1, ArrayList gene2, Trail p)
        {
            // Splice the new genes together
            phenotype = new ArrayList(gene1);
            phenotype.AddRange(gene2);
            trailPath = p;
        }

        public int CompareTo(Object obj)  
        {
            // Define a method to compare sinners to each other.
            //  -- For reproductive purposes. 

            Genotype g1 = (Genotype)obj;
            
            if (this.likelihood > g1.likelihood)
                return -1;
            else if (this.likelihood == g1.likelihood)
                return 0;
            else
                return 1;
         }

        public void printGene()
        {
            Console.WriteLine("Gene:\tFitness: {0}\tLocation: {1}\tLikelihood: {2:p}", fitness, location, Math.Round(likelihood,2));
            Console.WriteLine("Solution:");
            System.Collections.IEnumerator myEnumerator = phenotype.GetEnumerator();
            while (myEnumerator.MoveNext())
                Console.Write("{0} ", myEnumerator.Current);
            Console.WriteLine();
        }

        public ArrayList retrievePhenotype(int split, bool first)
        {
            // Do the brain surgery here. 
            //  1. Take a machette and chop the gene into two parts.

            if (first)
                return phenotype.GetRange(0, split);
            else
                return phenotype.GetRange(split, (phenotype.Count-split));
        }

        public void mutate(Random r)
        {
            int startPoint = r.Next(0, phenotype.Count - 1);    // Choose a start point from which to apply the mutation
            int freq = r.Next(1, 5);                            // Choose how often we're going to swap genes

            for (int idx = startPoint; idx <= phenotype.Count - 1; idx += freq)
            {
                //int swap = r.Next(1, 2); // Choose if we're going to swap or introduce a new random variable

                //if (swap == 1)
                //{ // Swap locations
                //    int other = r.Next(0, phenotype.Count-1);
                //    int temp = (int)phenotype[other];
                //    phenotype[other] = phenotype[idx];
                //    phenotype[idx] = temp;
                //}
                //else
                //{   // New random mutation
                    phenotype[idx] = r.Next(1, 100); 
                //}
            }
            //Console.WriteLine("Mutated gene");
        }

        public void calcLikelihood(int pSize, int avgFit)
        {
            // Calculate the fitness of the sinner. 
            //  -- If guy considering girl
            //      - 95% weight to girl.looks()
            //      - 4% to girl.smarts()
            //      - 1% to girl.otherGreatAttributes()
            //
            //  -- If girl considering guy
            //      - 60% to guy.hasMoney()
            //      - 20% to girl.currentMoodSwing()
            //      - 15% to guy.looks()
            //      - 4% to guy.smarts()
            //      - 1% to guy.otherGreatAttributes()

            likelihood = ((Convert.ToDouble(1) / Convert.ToDouble(pSize)) * (Convert.ToDouble(fitness) / Convert.ToDouble(avgFit)));
            if (likelihood == 0) likelihood = 0.01;
        }

        private void resetPhenotype()
        {
            fitness = 1;        // start square is first square of path
            location = 91;      // start bottom left
            lastLocation = 91;
            likelihood = 0;
        }

        public int Explore(int moves, Random r)
        {
            // Let the sinner wander around, who knows maybe something good will happen.

            // Directions: Up/Down & Left Right (Assuming facing up towards 1 from square 91)
            // Exceptions: SNum % 10 == 1 {Right, Up Down }
            //             SNum % 10 == 0 {Left, Up, Down }
            //             SNum <= 10 {Right, Left, Down }
            //             SNum >= 91 {Right, Left, Up }

            // Calculate where you can go
            //   - If rule in phenotype is found, follow it.
            //   - Else choose random direction. 
            //   - Update fitness

            resetPhenotype();
            
            ArrayList goTo = new ArrayList();
            ArrayList visited = new ArrayList();
            visited.Add(91);

            for (int c = 1; c <= moves; c++)
            {
                int newMove = 0;

                goTo.Add(location - 10);
                goTo.Add(location + 10);
                goTo.Add(location - 1);
                goTo.Add(location + 1);

                // Remove impossible border moves
                if (location % 10 == 1) goTo.Remove(location - 1);
                if (location % 10 == 0) goTo.Remove(location + 1);
                if (location <= 10) goTo.Remove(location - 10);
                if (location >= 91) goTo.Remove(location + 10);

                // Remove previous location, aka dont go back from where you just came from.
                // ASSERT(Assumption.. not a total idiot)
                goTo.Remove(lastLocation);

                // Check for phenotype rule
                for (int idx = 0; idx < goTo.Count; idx++)
                {
                    if (phenotype.Contains(goTo[idx]))
                    {
                        newMove = (int)goTo[idx];
                        // Check/Update Fitness
                        //   - If it's on path and visited !first! time increase fitness
                        if (trailPath.checkTrail(newMove) && !visited.Contains(newMove))
                            fitness++;
                        break;
                    }
                }

                // Didn't find anything in phenotype, pick random direction
                if (newMove == 0) newMove = (int)goTo[r.Next(0, goTo.Count-1)];
        
                lastLocation = location;
                location = newMove;
                visited.Add(lastLocation);
                goTo.Clear();
            }
            return fitness;
        } // Explore
    }
}